"""
Grokking the provided code will get us started.

  >>> testing.grok(__name__)

IStorageHandler provides an easy way to store items in annotations.
The first step is simple to instanciate a common persistent object.
Adapting it to IStorageHandler provides some convenient methods to
store, retrieve and delete annotation items.

  >>> item = ContentType()
  >>> item.id = 'item'
  >>> handler = IStorageHandler(item)
  >>> handler is not None
  True
  >>> IStorageHandler.providedBy(handler)
  True
  >>> IStorage.providedBy(handler.storage)
  True

As we can see, getting an annotation container is a matter of an adaptation.
Then, we can verify if the container has been effectively created.

  >>> from zope.annotation.interfaces import IAnnotations
  >>> annoted = IAnnotations(item)
  >>> 'troll.storage.tests' in annoted
  True
  >>> IStorage.providedBy(annoted['troll.storage.tests'])
  True
  >>> handler.storage is annoted['troll.storage.tests']
  True
  >>> len(annoted['troll.storage.tests'])
  0

Now, let's store and item in the annotation container. Storage and retrieval
are very straightforward.

  >>> foo = StoredItem('bar')
  >>> handler.store(foo)
  True
  >>> len(handler.storage)
  1
  >>> obj = handler.retrieve('bar')
  >>> isinstance(obj, StoredItem)
  True
  >>> obj.aq_parent is item
  True

Storage is non-destructive. If the entry already exists, it's not overriden.
Delete can be used to handle the destruction of the stored items.

  >>> usurper = StoredItem('bar')
  >>> handler.store(usurper)
  False
  >>> handler.delete('bar')
  True
  >>> handler.delete('bar')
  False
  >>> len(handler.storage)
  0
  >>> handler.store(usurper)
  True

Errors are raised if something went *very* wrong. Most of the classical errors
are already caught : non existing object on retrieval or deletion, etc.
Still, if you don't respect the storage integrity policy, you get troubles.

  >>> rogue = SimpleItem()
  >>> rogue.name = 'sneaky'
  >>> handler.store(rogue)
  Traceback (most recent call last):
        ...
  InvalidItemType: Trying to store a non IStorageItem object.
  >>> from zope.interface import directlyProvides
  >>> directlyProvides(rogue, IStorageItem)
  >>> handler.store(rogue)
  True

IStoredItem objects can be published. You can register views for them,
it works perfectly.

  >>> from zope.publisher.browser import TestRequest
  >>> from zope.component import getMultiAdapter
  >>> publishable = handler.retrieve('sneaky')
  >>> view = getMultiAdapter((publishable, TestRequest()), name='simple_view')
  >>> view.update()
  >>> view.render()
  '<h1>My name is sneaky and my parent is item</h1>'

"""

from five import grok
import five.grok.testing as testing
from Acquisition import Explicit
from OFS.SimpleItem import SimpleItem
from zope.interface import implements
from troll.storage.base import BaseStorageHandler
from troll.storage.descriptors import AnnotationContainerProperty
from troll.storage.interfaces import IStorageHandler, IStorage, IStorageItem
from zope.annotation.interfaces import IAttributeAnnotatable


class ContentType(SimpleItem):
    implements(IAttributeAnnotatable)


class StoredItem(Explicit):
    implements(IStorageItem)

    def __init__(self, name):
        self.name = name
        

class BaseStorageHandler(BaseStorageHandler):
    grok.context(ContentType)

    storage = AnnotationContainerProperty(
        IStorageHandler['storage'],
        ns='troll.storage.tests'
        )


class StoredItemView(grok.View):
    grok.name('simple_view')
    grok.context(IStorageItem)

    def render(self):
        return '<h1>My name is %s and my parent is %s</h1>' % (
            self.context.name, self.context.aq_inner.aq_parent.id
            )



# -*- coding: utf-8 -*-

try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)


from base import BaseStorageHandler
from interfaces import IStorage, IStorageItem, IStorageHandler
from descriptors import AnnotationContainerProperty



# -*- coding: utf-8 -*-

import unittest
from zope.component import testing
from zope.testing import doctestunit, doctest


import five.grok
import Products.Five
import troll.storage
import zope.annotation
from Products.Five import zcml


def setUp(test=None):
    testing.setUp(test)
    zcml.load_config('meta.zcml', package=Products.Five)
    zcml.load_config('configure.zcml', package=Products.Five)
    zcml.load_config('meta.zcml', package=five.grok)
    zcml.load_config('configure.zcml', package=five.grok)
    zcml.load_config('configure.zcml', package=zope.annotation)
    zcml.load_config('configure.zcml', package=troll.storage.tests)


def test_suite():
    return unittest.TestSuite([

        doctestunit.DocTestSuite(
            module='troll.storage.tests.annotation',
            setUp=setUp, tearDown=testing.tearDown),

        ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-

from interfaces import IStorage
from zope.interface import implements
from zope.app.container.btree import BTreeContainer


class GenericStorage(BTreeContainer):
    """A Generic BTree Container that implements IStorage.
    """
    implements(IStorage)
    



# -*- coding: utf-8 -*-

from five import grok
from descriptors import AnnotationContainerProperty
from interfaces import IStorageHandler, IStorageItem
from zope.app.container.interfaces import InvalidItemType


class BaseStorageHandler(grok.Adapter):
    """A very generic storage system
    It will store the items in a dictionnary-ish container.
    Therefore, keys are unique and can be of any hashable type.
    """
    grok.implements(IStorageHandler)
    grok.baseclass()

    @property
    def storage(self):
        return NotImplementedError(
            "Your storage handler has to provide its own storage."
            )

    def store(self, obj):
        if not IStorageItem.providedBy(obj):
            raise InvalidItemType(
                "Trying to store a non IStorageItem object."
                )
        key = obj.name
        if key not in self.storage.keys():
            try:
                self.storage[key] = obj
                return True
            except InvalidItemType:
                return False
        return False

    def retrieve(self, key):
        item = self.storage.get(key, None)
        if item is None:
            return None
        return self.storage[key].__of__(self.context)
            
    def delete(self, key):
        if key not in self.storage.keys():
            return False
        del self.storage[key]
        return True
        
    def __getattr__(self, name):
        return self.retrieve(name)



# -*- coding: utf-8 -*-

from Acquisition import aq_inner
from container import GenericStorage
from interfaces import IStorageItem
from zope.annotation.interfaces import IAnnotations

_marker = object()

class AnnotationContainerProperty(object):
    """Handles a container written in annotations.
    """
    def __init__(self, field, ns=u"troll.storage", name=None):
        self._name = name or field.__name__
        self._namespace = ns
        self.__field = field


    def __get__(self, inst, klass):
        field = self.__field.bind(inst)
        annotations = IAnnotations(aq_inner(inst.context))
        own_storage = annotations.get(self._namespace, _marker)
        if own_storage is _marker:
            own_storage = annotations[self._namespace] = GenericStorage()
        return own_storage
        

    def __set__(self, inst, value):
        field = self.__field.bind(inst)
        if field.readonly:
            raise ValueError(self._name, 'field is readonly')
        annotations = IAnnotations(aq_inner(inst.context))
        annotations[self._namespace] = value


    def __getattr__(self, name):
        return getattr(self.__field, name)



# -*- coding: utf-8 -*-

from zope.schema import TextLine, Object
from zope.interface import Interface, Attribute
from zope.app.container.interfaces import IContainer
from zope.app.container.constraints import containers, contains   


class IStorageItem(Interface):
    """An item that can be stored in an IStorage.
    """    
    name = TextLine(
        title = u"name",
        default = u"",
        required = True
        )

    
class IStorage(IContainer):
    """Marker interface for useable containers.
    An IStorage can only contain a IStorageItem
    """
    contains(IStorageItem)
    

class IStorageHandler(Interface):
    """A storage item handles the persistence of a given item.
    A storage has three main actions : store, retrieve, delete.
    """
    storage = Object(
        title = u"Container",
        description = u"Contains IStorageItem items.",
        schema = IStorage,
        required = True,
        readonly = False
        )

    def store(self, obj):
        """Stores an object and returns True.
        Returns False is the object can't be stored.
        """

    def retrieve(self, key):
        """Retrieve the object according to the given key.
        The retrieving method depends of the storage type.
        If no object is found, None is returned.
        """

    def delete(self, key):
        """Deletes the object with the given oid.
        Returns True if success, False otherwise.
        """



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from os.path import join

name = 'troll.storage'
version = '0.4'
readme = open("README.txt").read()

setup(name = name,
      version = version,
      description = 'Troll abstract storage layer',
      long_description = readme[readme.find('\n\n'):],
      keywords = 'zope grok storage',
      author = 'Souheil Chelfouh',
      author_email = 'souheil@chelfouh.com',
      url = 'http://tracker.trollfot.org/',
      download_url = 'http://pypi.python.org/pypi/troll.storage',
      license = 'GPL',
      packages = find_packages(),
      namespace_packages = ['troll'],
      include_package_data = True,
      platforms = 'Any',
      zip_safe = True,
      test_suite = 'troll.storage.tests',
      install_requires=[
          'setuptools',
          'five.grok'
      ],
      classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Zope',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
      ],
)
